#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include "mmlp.h"
#include "audio.h"

char* mmlp_version = "1.2";

extern double atof(const char *);
struct _channel_status* channel_status;
int current_channel;
static FILE* outfp = NULL;
double volume_ratio = 1.0;
int input_from_stdin = 0;
static int nchannel = 8;
char* surround_parent_name = NULL;
int port_offset = 0;
int mono_sound = 0;
int mono_sound_size = 2;
static int parse_mml2(void);
static int parse_mml_command(void);
int dev_audio_out = 1;
int check_parse_only = 0;
static int sigint_flag = 0;
#ifdef AUDRIV
int audriv_audio_flag = 0;
#include "audriv.h"
#include "aenc.h"
#endif /* AUDRIV */


void quit(int status)
{
    close_parent_fd();
    if(outfp != NULL)
	fclose(outfp);
#ifdef AUDRIV
    if(audriv_audio_flag)
    {
	if(!sigint_flag)
	{
	    while(audriv_play_active())
		audriv_wait_play();
	}
	audriv_play_stop();
	audriv_free_audio();
    }
#endif /* AUDRIV */
    close_audio(outfp);
    exit(status);
}

void illegal_command(void)
{
    if(trace_mode && !is_beginning_of_line)
	fputc('\n', stderr);
    fprintf(stderr, "Illegal command in %d\n", lineno);
}

void* xmalloc(int nbyte)
{
    void* p;

    if((p = (void *)malloc(nbyte)) == NULL && nbyte > 0)
    {
	fprintf(stderr, "Memory exhausted.\n");
	quit(1);
    }

    return p;
}

static void init_channel1(int i)
{
    channel_status[i].t = 120;
    channel_status[i].l = 4.0;
    channel_status[i].q = 8;
    channel_status[i].o = 4;
    channel_status[i].v = INIT_VOL;
    channel_status[i].neiro = 0;
    channel_status[i].m = 256;
    channel_status[i].s = 13;
#ifdef MIDI_CFG
    channel_status[i].tone = -1;
#endif /* MIDI_CFG */
}

static void init_channel(void)
{
    int i;

    channel_status = (struct _channel_status *)
	xmalloc(nchannel * sizeof(struct _channel_status));
    for(i = 0; i < nchannel; i++)
	init_channel1(i);
    current_channel = 0;
}

/* calclate play time [sec.] */
static INLINE double calc_playtime(double len)
{
    return (4.0 * 60.0) / channel_status[current_channel].t / len;
}

void ext_command(const char* cmd)
{
    if(strncmp(cmd, "PLAY", 4) == 0)
    {
	cmd += 4;
	while(*cmd && (*cmd == ' ' || *cmd == '\n')) cmd++;
	if(*cmd == '0' || strncmp(cmd, "OFF", 3) == 0)
	{
	    check_parse_only = 1;
	    if(trace_mode)
		fprintf(stderr, "Play off\n");
	}
	else
	{
	    check_parse_only = 0;
	    if(trace_mode)
		fprintf(stderr, "Play on\n");
	}
    }
    else if(strncmp(cmd, "TRACE", 5) == 0)
    {
	cmd += 5;
	while(*cmd && (*cmd == ' ' || *cmd == '\n')) cmd++;
	if(*cmd == '0' || strncmp(cmd, "OFF", 3) == 0)
	{
	    if(trace_mode)
		fprintf(stderr, "Trace off\n");
	    trace_mode = 0;
	}
	else
	{
	    fprintf(stderr, "Trace on\n");
	    trace_mode = 1;
	}
    }
    else if(strncmp(cmd, "HZ", 2) == 0)
    {
	cmd += 2;
	while(*cmd && (*cmd == ' ' || *cmd == '\n')) cmd++;
	set_base_Hz(atof(cmd));
    }
    else if(strncmp(cmd, "END", 3) == 0)
    {
	quit(0);
    }
    else
    {
	fprintf(stderr, "Unknown extension command at line %d\n", lineno);
    }
}


static void parse_play_scale(void)
{
    double play_time;
    int scale;
    double len;
    double f;
    int first_scale;
    int q8_mode;
    int tai_mode;
    double more;

    if(token_type == PLAY_SCALE)
	first_scale = (channel_status[current_channel].o - 1) * 12 + token_val;
    else
	first_scale = token_val;

    scale = first_scale;
    play_time = 0;
    q8_mode = 0;

  scale_concat:
    len = channel_status[current_channel].l;

    gettoken();
    if(token_type == NUMBER)
    {
	len = (double)token_val;
	gettoken();
    }

    len = more = calc_playtime(len);

    if(token_type == OP)
    {
	while(token_val == '.')
	{
	    gettoken();
	    more /= 2;
	    len += more;
	}
    }

    if(token_type == OP && token_val == '&')
    {
	gettoken();
	tai_mode = 1;
    }
    else
    {
	tai_mode = 0;
    }

    f = NtoHz(scale);
    play_time += len;

    if(tai_mode)
    {
	if(token_type == PLAY_SCALE ||
	   token_type == PLAY_SCALEN)
	{
	    if(token_type == PLAY_SCALE)
		scale
		    = (channel_status[current_channel].o - 1) * 12 + token_val;
	    else
		scale = token_val;
	    if(scale == first_scale)
		goto scale_concat;
	}
	q8_mode = 1;
    }

    make_play_wave(current_channel, play_time, f, q8_mode);
}

static void parse_mml_R(void)
{
    /* R[0-9]*\.* */
    double len, more;

    len = channel_status[current_channel].l;
    gettoken();
    if(token_type == NUMBER)
    {
	len = (double)token_val;
	gettoken();
    }
    len = more = calc_playtime(len);

    if(token_type == OP)
    {
	while(token_val == '.')
	{
	    gettoken();
	    more /= 2;
	    len += more;
	}
    }

    make_rest_wave(len);
}


static void parse_mml_L(void)
{
    double len;
    int channel = current_channel;

    gettoken();

    if(token_type != NUMBER)
    {
	channel_status[channel].l = 4.0;
	return;
    }

    len = (double)token_val;
    gettoken();

    if(token_type == OP)
    {
	while(token_val == '.')
	{
	    gettoken();
	    len /= 1.5;
	}
    }

    channel_status[channel].l = len;
}

static void parse_mml_Q(void)
{
    int channel = current_channel;
    gettoken();

    if(token_type != NUMBER)
    {
	channel_status[channel].q = 8;
	return;
    }
    channel_status[channel].q = token_val;
    gettoken();
}

static void parse_mml_O(void)
{
    int channel = current_channel;
    gettoken();
    if(token_type != NUMBER)
    {
	channel_status[channel].o = 4;
	return;
    }
    channel_status[channel].o = token_val;
    gettoken();
}

static void parse_mml_T(void)
{
    int channel = current_channel;
    gettoken();
    if(token_type != NUMBER)
    {
	channel_status[channel].t = 120;
	return;
    }
    channel_status[channel].t = token_val;
    gettoken();
}

static void parse_mml_M(void)
{
    int channel = current_channel;
    gettoken();
    if(token_type != NUMBER)
    {
	channel_status[channel].m = 256;
	return;
    }
    channel_status[channel].m = token_val;
    gettoken();
}

static void parse_mml_S(void)
{
    int channel = current_channel;
    gettoken();
    if(token_type != NUMBER)
    {
	channel_status[channel].s = 13;
	return;
    }
    channel_status[channel].s = token_val;
    gettoken();
}

static void rset_volume(int n)
{
    channel_status[current_channel].v += n / 15.0 * CHANNEL_VMAX;
}

static void parse_mml_atto(void)
{
    int channel = current_channel;
    gettoken();

    switch(token_type)
    {
      case NUMBER:
#ifdef MIDI_CFG
	channel_status[channel].tone  = token_val;
#else
	channel_status[channel].neiro = token_val;
#endif /* MIDI_CFG */
	gettoken();
	break;
      case COMMAND:
	if(token_val == '>')
	{
	    gettoken();
	    if(token_type == NUMBER)
	    {
		rset_volume(token_val);
		gettoken();
	    }
	    else
		rset_volume(1);
	    return;
	}
	else if(token_val == '<')
	{
	    gettoken();
	    if(token_type == NUMBER)
	    {
		rset_volume(-token_val);
		gettoken();
	    }
	    else
		rset_volume(-1);
	    return;
	}

	/* through */
      default:
#ifdef MIDI_CFG
	channel_status[channel].tone  = -1;
#else
	channel_status[channel].neiro = 0;
#endif /* MIDI_CFG */
    }
}

static void parse_mml_neiro(void)
{
    int channel = current_channel;
    gettoken();

    switch(token_type)
    {
      case NUMBER:
	channel_status[channel].neiro = token_val;
#ifdef MIDI_CFG
	channel_status[channel].tone  = -1;
#endif /* MIDI_CFG */
	gettoken();
	break;
      default:
	channel_status[channel].neiro = 0;
    }
}

static void parse_mml_down_octave(void)
{
    int channel = current_channel;
    gettoken();

    if(channel_status[channel].o == 1)
    {
	illegal_command();
	return;
    }
    channel_status[channel].o--;
}

static void parse_mml_up_octave(void)
{
    int channel = current_channel;
    gettoken();

    if(channel_status[channel].o == 9)
    {
	illegal_command();
	return;
    }
    channel_status[channel].o++;
}

static void parse_mml_V(void)
{
    int channel = current_channel;
    gettoken();
    if(token_type != NUMBER)
    {
	channel_status[channel].v = INIT_VOL;
    }
    else
    {
	channel_status[channel].v
	    = token_val / 15.0 * CHANNEL_VMAX;
	gettoken();
    }
}

static void parse_mml_v(void)
{
    int channel = current_channel;
    gettoken();
    if(token_type != NUMBER)
    {
	channel_status[channel].v = INIT_VOL;
    }
    else
    {
	channel_status[channel].v
	    = token_val / 127.0 * CHANNEL_VMAX;
	gettoken();
    }
}

static void parse_mml_Nthuplet(void)
{
    static int* scale_list = NULL;
    static int list_size = 0;
    static int list_max_size = 0;
    static const int default_list_size = 64;
    double len, play_time;
    int i;
    int nscale;

    nscale = 0;
    gettoken();
    while(1)
    {
	if(token_type == EOPhase ||
	   token_type == EOPart ||
	   token_type == EOF)
	{
	    if(trace_mode && !is_beginning_of_line)
		fputc('\n', stderr);
	    fprintf(stderr, "Parenthesis is mismatched in %d\n", lineno);
	    break;
	}
	if(token_type == COMMAND)
	{
	    if(token_val == '}')
	    {
		gettoken();
		break;
	    }
	    parse_mml_command();
	    continue;
	}

	if(token_type == START_LOOP)
	{
	    begin_loop(token_val);
	    gettoken();
	    continue;
	}

	if(token_type == END_LOOP)
	{
	    end_loop();
	    gettoken();
	    continue;
	}

	if(token_type != PLAY_SCALE &&
	   token_type != PLAY_SCALEN)
	{
	    gettoken();
	    continue;
	}

	if(list_size == list_max_size)
	{
	    if(list_max_size == 0)
	    {
		list_max_size = default_list_size;
		scale_list = (int *)xmalloc(list_max_size * sizeof(int));
	    }
	    else
	    {
		list_max_size *= 2;
		scale_list = (int *)realloc(scale_list,
					    list_max_size * sizeof(int));

		if(scale_list == NULL)
		{
		    fprintf(stderr, "Memory is exhausted.\n");
		    quit(1);
		}
	    }
	}
	if(token_type == PLAY_SCALEN)
	    scale_list[list_size] = token_val;
	else
	    scale_list[list_size]
		= (channel_status[current_channel].o - 1) * 12 + token_val;
	list_size++;
	nscale++;
	gettoken();
    }

    list_size -= nscale;

    if(nscale == 0)
	return;
    if(token_type == NUMBER)
    {
	len = (double)token_val;
	gettoken();
    }
    else
	len = channel_status[current_channel].l;

    if(token_type == OP)
    {
	while(token_val == '.')
	{
	    gettoken();
	    len /= 1.5;
	}
    }
    play_time = calc_playtime(len);

    for(i = 0; i < nscale; i++)
    {
	make_play_wave(current_channel, play_time / nscale,
		       NtoHz(scale_list[list_size + i]), 0);
    }
}


static int parse_mml_command(void)
{
    switch(token_val)
    {
      case 'V':
	parse_mml_V();
	break;
      case 'v':		/* @V */
	parse_mml_v();
	break;
      case 'L':
	parse_mml_L();
	break;
      case 'Q':
	parse_mml_Q();
	break;
      case 'O':
	parse_mml_O();
	break;
      case 'T':
	parse_mml_T();
	break;
      case 'R':
	parse_mml_R();
	break;
      case '@':
	parse_mml_atto();
	break;
      case 'Y':
	parse_mml_neiro();
	break;
      case 'M':
	parse_mml_M();
	break;
      case 'S':
	parse_mml_S();
	break;
      case '<':
	parse_mml_down_octave();
	break;
      case '>':
	parse_mml_up_octave();
	break;
      case '{':
	parse_mml_Nthuplet();
	break;
      default:
	illegal_command();
	gettoken();
	return 0;
    }
    return 1;
}

static int parse_mml2(void)
{
    switch(token_type)
    {
      case EOPhase:
      case EOPart:
      case EOF:
	return 2;
      case PLAY_SCALE:
      case PLAY_SCALEN:
	parse_play_scale();
	break;
      case COMMAND:
	return parse_mml_command();
      case START_LOOP:
	begin_loop(token_val);
	gettoken();
	break;
      case END_LOOP:
	end_loop();
	gettoken();
	break;
      default:
	illegal_command();
	gettoken();
	return 0;
    }
    return 1;
}

static void parse_mml1(void)
{
    int status;

    while(1)
    {
	status = parse_mml2();
	if(status == 2)
	    return;
    }
}

static void expand_channel(void)
{
    int i;

    channel_status = (struct _channel_status *)
	realloc(channel_status, 2 * nchannel * sizeof(struct _channel_status));
    if(channel_status == NULL)
    {
	fprintf(stderr, "Memory exhausted.\n");
	quit(1);
    }

    for(i = nchannel; i < nchannel * 2; i++)
	init_channel1(i);

    nchannel *= 2;
}	

static void parse_mml0(void)
{
    if(token_type == EOPhase)
    {
	current_channel = 0;
	gettoken();
	return;
    }

    while(1)
    {
	if(token_type == EOPart)
	{
	    current_channel++;
	    if(current_channel == nchannel)
		expand_channel();
	    gettoken();
	    start_position_play();
	    continue;
	}

	if(token_type == EOPhase || token_type == EOF)
	{
	    current_channel = 0;
	    gettoken();
	    break;
	}
	parse_mml1();
    }
}

static char* get_file_text(const char* filename)
{
    char* s;
    int fd;
    struct stat buf;
    int i;

    if((fd = open(filename, O_RDONLY)) == -1)
    {
	perror(filename);
	quit(1);
    }

    if(fstat(fd, &buf) == -1)
    {
	perror(filename);
	quit(1);
    }

    s = (char *)xmalloc(buf.st_size + 2);

    read(fd, s, buf.st_size);
    close(fd);

    if(s[buf.st_size - 1] != '\n')
    {
	s[buf.st_size] = '\n';
	s[buf.st_size + 1] = '\0';
    }
    else
    {
	s[buf.st_size] = '\0';
    }

    for(i = 0; i < buf.st_size; i++)
	s[i] = toupper(s[i]);

    return s;
}

FILE* write_open(const char* fname)
{
    FILE* fp;

    if(!strcmp(fname, "-"))
	return stdout;

#ifdef AUDRIV
    if(strcmp(fname, "/dev/audio") == 0)
    {
	fname = "/dev/null";
	audriv_audio_flag = 1;
    }
#endif /* AUDRIV */


    if((fp = fopen(fname, "w")) == NULL)
    {
	perror(fname);
	quit(1);
    }

    return fp;
}

char* get_env(const char* var)
{
    char* x;

    if((x = (char *)getenv(var)) == (char *)-1)
    {
	fprintf(stderr, "%s: Undefined evnironment.\n", var);
	exit(1);
    }

    return x;
}

int parse_mono_sound_size(char* s, int* x, int* y)
{
    char* p;

    p = s;

    if(!isdigit(*p))
	return 0;

    p++;
    while(isdigit(*p))
	p++;

    *x = atoi(s);
    if(*p != '/')
	return 1;

    p++;
    s = p;

    if(!isdigit(*p))
	return 0;

    p++;
    while(isdigit(*p))
	p++;

    *y = atoi(s);
    return 1;
}

static void usage(void)
{
    fprintf(stderr, "usage: mmlp [-Hz <base-Hz>] [-vol <volume>] [-s] [-v] [-h] [-mspart <1 or 2>] [filename [output-filename]]\n");
    fprintf(stderr, "usage: mmlp -sr [-Hz <base-Hz>] [-vol <volume>] [-port <port-number>] [-ms] hostname:filename [...]\n");
}

void sigint_handler()
{
    sigint_flag = 1;
    quit(0);
}

void main(int argc, char** argv)
{
    int i;
    char* outfname = "/dev/audio";
    char* infilename;
    int surround_mode = 0;
    int surround_child = 0;

    for(i = 1; i < argc; i++)
    {
	if(!strcmp(argv[i], "-Hz"))
	{
	    i++;
	    if(i == argc)
	    {
		usage();
		quit(1);
	    }
	    set_base_Hz(atof(argv[i]));
	}
	else if(!strcmp(argv[i], "-vol"))
	{
	    i++;
	    if(i == argc)
	    {
		usage();
		quit(1);
	    }
	    volume_ratio = atof(argv[i]);
	    if(volume_ratio < 0)
	    {
		usage();
		quit(1);
	    }
	}
	else if(!strcmp(argv[i], "-port"))
	{
	    i++;
	    if(i == argc)
	    {
		usage();
		quit(1);
	    }
	    port_offset = atoi(argv[i]);
	}
	else if(!strcmp(argv[i], "-s"))
	{
	    trace_mode = 0;
	}
	else if(!strcmp(argv[i], "-check"))
	{
	    check_parse_only = 1;
	}
	else if(!strcmp(argv[i], "-surround-child"))
	{
	    trace_mode = 0;
	    surround_child = 1;
	    i++;
	    if(argv[i] == NULL)
	    {
		fprintf(stderr, "mmlp: Systematic error: Bad argument of child prosess: %d\n", (int)getpid());
		exit(1);
	    }
	    surround_parent_name = argv[i];
	}
	else if(!strcmp(argv[i], "-ms"))
	{
	    ;
	}
	else if(!strcmp(argv[i], "-nms"))
	{
	    i++;
	    if(argv[i] == NULL)
	    {
		usage();
		exit(1);
	    }
	    mono_sound_size = atoi(argv[i]);
	    if(mono_sound_size <= 0)
	    {
		usage();
		exit(1);
	    }
	}
	else if(!strcmp(argv[i], "-mspart"))
	{
	    i++;
	    if(argv[i] == NULL)
	    {
		usage();
		exit(1);
	    }

	    if(parse_mono_sound_size(argv[i],
				     &mono_sound,
				     &mono_sound_size) == 0 ||
	       mono_sound_size == 0 ||
	       mono_sound == 0 ||
	       mono_sound > mono_sound_size)
	    {
		usage();
		exit(1);
	    }
	}

	else if(!strcmp(argv[i], "-sr"))
	{
	    surround_mode = 1;
	}
	else if(!strcmp(argv[i], "-h"))
	{
	    usage();
	    quit(0);
	}
	else if(!strcmp(argv[i], "-v"))
	{
	    fprintf(stderr, "mmlp version %s\n", mmlp_version);
	    quit(0);
	}
	else if(argv[i][0] == '-' &&
		argv[i][1] != '\0')
	{
	    usage();
	    quit(1);
	}
	else
	{
	    break;
	}
    }

    if(surround_mode)
    {
	if(i == argc)
	{
	    usage();
	    quit(1);
	}
	exec_remote_clild(argv, argc, i);
    }

    if(i == argc)
	infilename = "-";
    else
    {
	infilename = argv[i];
	i++;
    }

    if(!strcmp(infilename, "-") ||
       !strcmp(infilename, "/dev/tty"))
	input_from_stdin = 1;
    else
	input_from_stdin = 0;

    if(!input_from_stdin)
    {
	mml_text = get_file_text(infilename);
	add_mmltext(mml_text);
    }

    if(i < argc)
    {
	outfname = argv[i];
	if(!strcmp(outfname, "/dev/null"))
	    dev_audio_out = 0;
    }
    outfp = write_open(outfname);

#ifdef AUDRIV
    if(audriv_audio_flag)
    {
	if(!audriv_setup_audio())
	{
	    fprintf(stderr, "%s: %s\n", argv[0], audriv_errmsg);
	    exit(0);
	}
	audriv_set_noblock_write(False);
	audriv_set_play_encoding(AENC_G711_ULAW);
	audriv_play_open();

    }
#endif /* AUDRIV */

    if(trace_mode)
	fprintf(stderr, "mmlp - ver. %s, Copyright (c) 1996, Masanao Izumo.\n",
		mmlp_version);

    init_channel();

    signal(SIGINT, sigint_handler);

    if(!input_from_stdin)
    {
	gettoken();
	while(token_type != EOF)
	{
	    clear_play_buffer();
	    parse_mml0();
	    play(outfp);
	}
    }
    else
    {
	static char buff[BUFSIZ];
	int i;
	char* p;

	if(surround_parent_name != NULL)
	{
	    fprintf(stderr, "Can't play the mml from stdin.\n");
	    exit(1);
	}

	p = NULL;
	while(fgets(buff, BUFSIZ, stdin) != NULL)
	{
	    int len;

	    len = strlen(buff);
	    for(i = 0; i < len; i++)
		buff[i] = toupper(buff[i]);

	    if(p == NULL)
		p = mml_text = strdup(buff);
	    else
		p = strdup(buff);

	    if(p == NULL)
	    {
		fprintf(stderr, "Memory exhausted.\n");
		quit(1);
	    }
	    add_mmltext(p);
	    if(len > 2 &&
	       p[len - 2] == '\\' &&
	       p[len - 1] == '\n')
		continue;

	    is_beginning_of_line = 1;
	    gettoken();
	    while(token_type != EOF)
	    {
		clear_play_buffer();
		parse_mml0();
		play(outfp);
	    }
	    p = NULL;
	}
    }

    if(trace_mode && !is_beginning_of_line)
	fputc('\n', stderr);

    quit(0);
}
